from dweather_client import utils, http_client

def test_snap_to_grid_chirps():
    heads = http_client.get_heads()
    chirps_metadata = http_client.get_metadata(heads['chirps_05-daily'])
    lat, lon = utils.snap_to_grid(41.1842, -75.11, chirps_metadata)
    assert lat == 41.175
    assert lon == -75.125

def test_snap_to_grid_prism():
    heads = http_client.get_heads()
    prism_metadata = http_client.get_metadata(heads['prism_precip-daily'])
    lat, lon = utils.snap_to_grid(39.8398, -104.193, prism_metadata)
    assert lat == 39.833
    assert lon == -104.208

def test_snap_to_grid_cpc_global_daily():
    # tests snap_to_grid with ideosyncratic cpc lat/lon format
    heads = http_client.get_heads()
    prism_metadata = http_client.get_metadata(heads['cpc_global-daily'])
    lat, lon = utils.snap_to_grid(69.754, 330.759, prism_metadata)
    assert lat == 69.750
    assert lon == 330.750

def test_get_n_closest_station_ids():
    heads = http_client.get_heads()
    ghcnd_metadata = http_client.get_metadata(heads['ghcnd'])

    # get the 20 closest station ids to a spot in Kentucky
    ids = utils.get_n_closest_station_ids(37, -85, ghcnd_metadata, 20)
    print(ids)


"""
Basic functions for getting data from a dWeather gateway via https.
"""
import requests, datetime, io, gzip, json, csv
from dweather_client.ipfs_errors import *
from dweather_client.utils import listify_period, celcius_to_fahrenheit
import dweather_client.ipfs_datasets
from collections import deque

MM_TO_INCHES = 0.0393701
RAINFALL_PRECISION = 5
GATEWAY_URL = 'https://gateway.arbolmarket.com'


def get_heads(url=GATEWAY_URL):
    """
    Get heads.json for a given IPFS gateway.
    Args:
        url (str): base url of the IPFS gateway url
    Returns (example heads.json):
        {
            'chirps_05-daily': 'Qm...',
            'chirps_05-monthly': 'Qm...',
            'chirps_25-daily': 'Qm...',
            'chirps_25-monthly': 'Qm...',
            'cpc_us-daily': 'Qm...',
            'cpc_us-monthly': 'Qm...'
        }
    """
    hashes_url = url + "/climate/hashes/heads.json"
    r = requests.get(hashes_url)
    r.raise_for_status()
    return r.json()


def get_metadata(hash_str, url=GATEWAY_URL):
    """
    Get the metadata file for a given hash.
    Args:
        url (str): the url of the IPFS server
        hash_str (str): the hash of the ipfs dataset
    Returns (example metadata.json):
    
        {
            'date range': [
                '1981/01/01',
                '2019/07/31'
            ],
            'entry delimiter': ',',
            'latitude range': [
                -49.975, 49.975
            ],
            'longitude range': [
                -179.975, 179.975]
            ,
            'name': 'CHIRPS .05 Daily Full Set Uncompressed',
            'period': 'daily',
            'precision': 0.01,
            'resolution': 0.05,
            'unit of measurement': 'mm',
            'year delimiter': '\n'
        }
    """
    metadata_url = "%s/ipfs/%s/metadata.json" % (url, hash_str)
    r = requests.get(metadata_url)
    r.raise_for_status()
    return r.json()


def traverse_ll(head):
    release_itr = head
    release_ll = deque()
    while True:
        release_ll.appendleft(release_itr)
        prev_release = get_metadata(release_itr)['previous hash']
        if prev_release != None:
            release_itr = prev_release
        else:
            return release_ll


def get_hurricane_release_dict(release_hash, url=GATEWAY_URL):
    url = "%s/ipfs/%s/history.json.gz" % (url, release_hash)
    resp = requests.get(url)
    resp.raise_for_status()
    with gzip.GzipFile(fileobj=io.BytesIO(resp.content)) as zip_data:
        return json.loads(zip_data.read().decode("utf-8"))


def get_hurricane_dict(head=get_heads()['atcf_btk-seasonal']):
    """
    Get a hurricane dictionary for the atcf_btk-seasonal dataset. 

    To get a unique value to query the dict by storm, use BASIN + CY + the year
    part of the HOUR value. BASIN is the ocean, CY is the storm index, and
    the year is needed as well because the storm index resets every year.

    Note that there will be multiple readings with the same HOUR value,
    as readings are taken more than once per hour and then rounded to the nearest
    hour before posting. 
    """
    heads = get_heads()
    hurr_head = heads['atcf_btk-seasonal']
    release_ll = traverse_ll(hurr_head)
    hurr_dict = {}
    for release_hash in release_ll:
        release_content = get_hurricane_release_dict(release_hash)
        try:
            hurr_dict['features'].append(release_content['features'])
        except KeyError:
            hurr_dict.update(release_content)
    return hurr_dict


def get_station_csv(station_id, url=GATEWAY_URL):
    """
    Retrieve the contents of a station data csv file.
    Args:
        station_id (str): the id of the weather station
    returns:
        the contents of the station csv file as a string
    """
    all_hashes = get_heads()
    dataset_hash = all_hashes["ghcnd"]
    dataset_url = "%s/ipfs/%s/%s.csv.gz" % (url, dataset_hash, str(station_id))
    r = requests.get(dataset_url)
    r.raise_for_status()
    with gzip.GzipFile(fileobj=io.BytesIO(r.content)) as zip_data:
        return zip_data.read().decode("utf-8")


def parse_station_temps_as_dict(csv_text, use_fahrenheit=True):
    """
    Parse a station CSV file and get column values
    Will automatically index by date
    Args:
        csv_text (str): the GHCND station csv text
        use_fahrenheight (bool): if true use deg F, otherwise degrees C
    Returns:
        tuple:
            dict of datetime.date: float temperature highs
            dict of datetime.date: float temperature lows
    """
    #csv_text = get_station_csv(station_id)
    reader = csv.reader(csv_text.split())
    column_names = next(reader)
    date_col = column_names.index('DATE')
    tmax_col = column_names.index('TMAX')
    tmin_col = column_names.index('TMIN')
    tmins = {}
    tmaxs = {}
    for row in reader:
        # data is in tenths of a degree C
        if row[tmin_col] == '' or row[tmax_col] == '':
            continue
        tmax = float(row[tmax_col])/10.0
        tmin = float(row[tmin_col])/10.0
        if use_fahrenheit:
            tmax = celcius_to_fahrenheit(tmax)
            tmin = celcius_to_fahrenheit(tmin)
        tmaxs[datetime.datetime.strptime(row[date_col], "%Y-%m-%d").date()] = tmax
        tmins[datetime.datetime.strptime(row[date_col], "%Y-%m-%d").date()] = tmin

    return tmins, tmaxs


def get_station_by_wmo_id(wmo_id):
    pass


def get_station_by_airport_code(code):
    pass


def get_hash_cell(hash_str, coord_str, url=GATEWAY_URL):
    dataset_url = '%s/ipfs/%s/%s' % (url, hash_str, coord_str)
    r = requests.get(dataset_url)
    r.raise_for_status()
    return r.text


def get_zipped_hash_cell(hash_str, coord_str, url=GATEWAY_URL):
    """
    Read a text file on the ipfs server compressed with gzip.
    Args:
        url (str): the url of the ipfs server
        hash_str (str): the hash of the dataset
        coord_str (str): the text file coordinate name e.g. 45.000_-96.000
    Returns:
        the contents of the file as a string
    """
    dataset_url = '%s/ipfs/%s/%s.gz' % (url, hash_str, coord_str)
    r = requests.get(dataset_url)
    r.raise_for_status()
    with gzip.GzipFile(fileobj=io.BytesIO(r.content)) as zip_data:
        return zip_data.read().decode("utf-8")


def get_dataset_cell(lat, lon, dataset_revision):
    """ 
    Retrieve the text of a grid cell data file for a given lat lon and dataset.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        A tuple (json, str) of the dataset metadata file and the grid cell data text
    Raises: 
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
    """
    all_hashes = get_heads()
    if dataset_revision in all_hashes:
        dataset_hash = all_hashes[dataset_revision]
    else:
        raise DatasetError('{} not found on server'.format(dataset_revision))

    metadata = get_metadata(dataset_hash)
    min_lat, max_lat = sorted(metadata["latitude range"])
    min_lon, max_lon = sorted(metadata["longitude range"])
    if lat < min_lat or lat > max_lat:
        raise InputOutOfRangeError("Latitude {} out of dataset revision range [{:.3f}, {:.3f}] for {}".format(lat, min_lat, max_lat, dataset_revision))
    if  lon < min_lon or lon > max_lon:
        raise InputOutOfRangeError("Longitude {} out of dataset revision range [{:.3f}, {:.3f}] for {}".format(lon, min_lon, max_lon, dataset_revision))
    coord_str = "{:.3f}_{:.3f}".format(lat,lon)
    try:
        if "compression" in metadata and metadata["compression"] == "gzip":
            text_data = get_zipped_hash_cell(dataset_hash, coord_str)
        else:
            text_data = get_hash_cell(dataset_hash, coord_str)
        return metadata, text_data
    except requests.exceptions.RequestException as e:
        raise CoordinateNotFoundError('Coordinate ({}, {}) not found  on ipfs in dataset revision {}'.format(lat, lon, dataset_revision))


def get_rainfall_dict(lat, lon, dataset_revision, return_metadata=False):
    """ 
    Build a dict of rainfall data for a given grid cell.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        a dict ({datetime.date: float}) of datetime dates and the corresponding rainfall in mm for that date
    Raises:
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as rainfall data
    """
    metadata, rainfall_text = get_dataset_cell(lat, lon, dataset_revision)
    dataset_start_date = datetime.datetime.strptime(metadata['date range'][0], "%Y/%m/%d").date()
    dataset_end_date = datetime.datetime.strptime(metadata['date range'][1], "%Y/%m/%d").date()
    timedelta = dataset_end_date - dataset_start_date
    days_in_record = timedelta.days + 1 # we have both the start and end date in the dataset so its the difference + 1
    day_strs = rainfall_text.replace(',', ' ').split()
    if (len(day_strs) != days_in_record):
        raise DataMalformedError ("Number of days in data file does not match the provided metadata")
    rainfall_dict = {}
    for i in range(days_in_record):
        if day_strs[i] == metadata["missing value"]:
            rainfall_dict[dataset_start_date + datetime.timedelta(days=i)] = None
        else:
            rainfall_dict[dataset_start_date + datetime.timedelta(days=i)] = float(day_strs[i])
    if return_metadata:
        return metadata, rainfall_dict
    else:
        return rainfall_dict


def get_rev_rainfall_dict(lat, lon, dataset, desired_end_date, latest_rev):
    """
    Build a dictionary of rainfall data. Include as much of the most accurate, final data as possible. Start by buidling from the most accurate data,
    then keep appending data from more recent/less accurate versions of the dataset until we run out or reach the end date.

    This will not throw an error if there are no revisions with data available, it will simply return what is available.
    Args:
        lat (float): the grid cell latitude
        lon (float): the grid cell longitude
        dataset (str): the name of the dataset, e.g., "chirps_05-daily" on hashes.json
        desired_end_date (datetime.date): the last day of data needed.
        latest_rev (str): the least accurate revision of the dataset that is considered final
    Returns:
        tuple:
            a dict ({datetime.date: float}) of datetime dates and the corresponding rainfall in mm for that date
            bool is_final: if all data up to desired end date is final, this will be true
    """
    all_rainfall = {}
    is_final = True

    # Build the rainfall from the most accurate revision of the dataset to the least
    for dataset_revision in dweather_client.ipfs_datasets.datasets[dataset]:
        additional_rainfall = get_rainfall_dict(lat, lon, dataset_revision)
        all_dates = list(all_rainfall) + list(additional_rainfall)
        # This method of dict comprehension preserves the order of the dict
        all_rainfall = {date: all_rainfall[date] if date in all_rainfall else additional_rainfall[date] for date in all_dates}
        # stop when we have the desired end date in the dataset
        if desired_end_date in all_rainfall:
            return all_rainfall, is_final
        # data is no longer final after we pass the specified version
        if dataset_revision == latest_rev:
            is_final = False

    # If we don't reach the desired dataset, return all data.
    return all_rainfall, is_final


def get_temperature_dict(lat, lon, dataset_revision, return_metadata=False):
    """
    Build a dict of temperature data for a given grid cell.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        tuple (highs, lows) of dicts
        highs: dict ({datetime.date: float}) of datetime dates and the corresponding high temperature in degress F
        lows: dict ({datetime.date: float}) of datetime dates and the corresponding low temperature in degress F
    Raises:
        DatasetError: If no matching dataset_revision found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset_revision range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as temperature data
    """
    metadata, temp_text = get_dataset_cell(lat, lon, dataset_revision)
    dataset_start_date = datetime.datetime.strptime(metadata['date range'][0], "%Y/%m/%d").date()
    dataset_end_date = datetime.datetime.strptime(metadata['date range'][1], "%Y/%m/%d").date()
    timedelta = dataset_end_date - dataset_start_date
    days_in_record = timedelta.days + 1 # we have both the start and end date in the dataset_revision so its the difference + 1
    day_strs = temp_text.replace(',', ' ').split()
    if (len(day_strs) != days_in_record):
        raise DataMalformedError ("Number of days in data file does not match the provided metadata")
    highs = {}
    lows = {}
    for i in range(days_in_record):
        low, high = map(float, day_strs[i].split('/'))
        date_iter = dataset_start_date + datetime.timedelta(days=i)
        highs[date_iter] = high
        lows[date_iter] = low
    if return_metadata:
        return metadata, highs, lows
    else:
        return highs, lows


def get_rev_temperature_dict(lat, lon, dataset, desired_end_date, latest_rev):
    """
    Build a dictionary of rainfall data. Include as much final data as possible. If the desired end date
    is not in the final dataset, append as much prelim as possible.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
        dataset (str): the dataset name as on hashes.json
        desired_end_date (datetime.date): don't include prelim data after this point if not needed
        latest_rev (str): The least accurate revision that is still considered 'final'
    returns:
        tuple (highs, lows) of dicts and a bool
        highs: dict ({datetime.date: float}) of datetime dates and the corresponding high temperature in degress F
        lows: dict ({datetime.date: float}) of datetime dates and the corresponding low temperature in degress F
        is_final: True if all data is from final dataset, false if prelim included
    """
    highs = {}
    lows = {}
    is_final = True

    # Build the data from the most accurate version of the dataset to the least
    for dataset_revision in dweather_client.ipfs_datasets.datasets[dataset]:
        additional_highs, additional_lows = get_temperature_dict(lat, lon, dataset_revision)
        all_dates = list(highs) + list(additional_highs)    
        highs = {date: highs[date] if date in highs else additional_highs[date] for date in all_dates}
        lows = {date: lows[date] if date in lows else additional_lows[date] for date in all_dates}
        # Stop early if we have the end date
        if desired_end_date in highs:
            return highs, lows, is_final

        # data is no longer final after we pass the specified version
        if dataset_revision == latest_rev:
            is_final = False

    # If we don't reach the desired dataset, return all data.
    return highs, lows, is_final


def get_rev_tagged_temperature_dict(lat, lon, dataset, desired_end_date=None):
    ''' Build temps with a revision tag by each date
    Args:
        lat (float): the grid cell latitude
        lon (float): the grid cell longitude
        dataset (str): the dataset name in ipfs
        desired_end_date (datetime.date): stop early if we get this end date
    returns
        highs: dict with keys of dates, values are tuple (temperature, revision tag) for that date
        lows: dict with keys of dates, values are tuple (temperature, revision tag) for that date
    '''
    highs = {}
    lows = {}

    # Build the data from the most accurate version of the dataset to the least
    for dataset_version in dweather_client.ipfs_datasets.datasets[dataset]:
        additional_highs, additional_lows = get_temperature_dict(lat, lon, dataset_version)
        all_dates = list(highs) + list(additional_highs)    
        highs = {date: highs[date] if date in highs else (additional_highs[date], dataset_version) for date in all_dates}
        lows = {date: lows[date] if date in lows else (additional_lows[date], dataset_version) for date in all_dates}
        # Stop early if we have the end date
        if desired_end_date in highs:
            return highs, lows

    # If we don't reach the desired dataset, return all data.
    return highs, lows






from dweather_client.df_arithmetic import cat_icao_stations, cat_n_closest_station_dfs
from dweather_client.http_client import get_heads, get_metadata

#def test_cat_icao_stations():
#    ghcnd_hash = get_heads()['ghcnd']
#    stations = cat_icao_stations(pin=False, force_hash=ghcnd_hash)

def test_cat_n_closest_station_dfs():
    ghcnd_hash = get_heads()['ghcnd'] 
    stations = cat_n_closest_station_dfs( \
        36, 
        -94.5, 
        10, 
        pin=False, 
        force_hash=ghcnd_hash
    )



from dweather_client import http_client, utils
from dweather_client.data_loader import GridCellDataLoader
import datetime

def test_get_revision_rainfall():
    dataset = "chirps_05"
    dataset_revision = "chirps_prelim_05-daily"
    coords = 41.175, -75.125
    loader = GridCellDataLoader(dataset, *coords, "rainfall")
    http_client_meta, http_client_dict = http_client.get_rainfall_dict(*coords, dataset_revision, return_metadata=True)
    loader_data = loader.get_revision(dataset_revision)
    loader_dict = loader_data["data"]
    loader_meta = loader_data["metadata"]
    assert http_client_dict == loader_dict
    assert http_client_meta == loader_meta


def test_get_revision_temperature():
    dataset = "prism_temp"
    dataset_revision = 'prism_temp-daily'
    coords=41.25, -77.75
    loader = GridCellDataLoader(dataset, *coords, "temperature")
    http_client_meta, http_client_highs, http_client_lows = http_client.get_temperature_dict(*coords, dataset_revision, return_metadata=True)
    loader_data = loader.get_revision(dataset_revision)
    loader_dict = loader_data["data"]
    loader_meta = loader_data["metadata"]
    assert http_client_highs == loader_dict["highs"]
    assert http_client_lows == loader_dict["lows"]
    assert http_client_meta == loader_meta

def test_multi_revision_dict_rainfall():
    dataset = "prism_precip"
    coords = 41.25, -77.75
    loader = GridCellDataLoader(dataset, *coords, "rainfall")
    loader_rain = loader.build_multi_revision_dict()
    http_rain, _ = http_client.get_rev_rainfall_dict(*coords, dataset, datetime.date(2020, 1, 1), "prism_rev_1_precip-daily")

    for date in http_rain:
        assert loader_rain[date][0] == http_rain[date]


def test_multi_revision_dict_temperature():
    dataset = "prism_temp"
    coords = 41.25, -77.75
    loader = GridCellDataLoader(dataset, *coords, "temperature")
    loader_temps = loader.build_multi_revision_dict()
    http_highs, http_lows = http_client.get_rev_tagged_temperature_dict(*coords, dataset, datetime.date(2020, 1, 1))

    for date in http_highs:
        assert loader_temps["highs"][date][0] == http_highs[date][0]
        assert loader_temps["lows"][date][0] == http_lows[date][0]

def test_single_instance_enforcement():
    chirps_args = "chirps_05", 41.625, -93.125, "rainfall"
    loader_a = GridCellDataLoader(*chirps_args)
    loader_b = GridCellDataLoader(*chirps_args)
    assert loader_a is loader_b
    prism_args = "prism_precip", 41.25, -77.75, "rainfall"
    loader_c = GridCellDataLoader(*prism_args)
    assert loader_a is not loader_c




class IPFSError(Exception):
    """Base class for exceptions in this module."""
    pass

class DatasetError(IPFSError):
    """Excpetion raised when a dataset cannot be found on ipfs"""
    pass

class InputOutOfRangeError(IPFSError):
    """Exception raised for lat/lon requested outside the valid range for a dataset"""
    pass

class CoordinateNotFoundError(IPFSError):
    """Exception raised when a lat/lon coordinate pair does not have a file on the server"""
    pass

class DataMalformedError(IPFSError):
    """Raised when a grid cell text file is unable to be parsed according to metadata"""
    pass


### How this file works
''' How to use this file
Each entry is a sorted list of versions of the dataset, from most accurate to least
'''

datasets = {
	"prism_precip" : [
		"prism_rev_6_precip-daily",
		"prism_rev_3_precip-daily",
		#"prism_rev_1_precip-daily"
	],
	"prism_temp" : [
		"prism_rev_6_temp-daily",
		"prism_rev_3_temp-daily",
		"prism_rev_1_temp-daily"
	],
	"chirps_05": [
		"chirps_05-daily",
		"chirps_prelim_05-daily"
	],
	"chirps_25": [
		"chirps_25-daily"
	],
	"cpc_global": [
		"cpc_global-daily"
	],
	"cpc_us": [
		"cpc_us-daily"
	],
	"cpc_temp": [
		"cpc_temp-daily"
	]
}



from dweather_client import http_client, utils
from datetime import date

def test_station():
    station = http_client.get_station_csv("ACW00011604")
    station_content = '"ACW00011604","1949-01-12","17.11667","-61.78333","10.1","ST JOHNS COOLIDGE FIELD, AC","    0","T,,X","    0",",,X","    0",",,X","  278",",,X","  194",",,X",,,,,,,,,,,"    1",",,X"'
    assert(station_content in station)



"""
Basic functions for getting data from a dWeather gateway via https if you prefer to work
in pandas dataframes rather than Python's built in types. A wrapper for http_client.
"""
from dweather_client.http_client import get_rainfall_dict, get_temperature_dict, get_station_csv
from dweather_client.ipfs_client import cat_station_csv
from dweather_client.df_utils import get_station_ids_with_icao
import pandas as pd
import io
import ipfshttpclient

def get_rainfall_df(lat, lon, dataset):
    """
    Get full daily rainfall time series from cpc, prism, or chirps in mm.
    
    return:
        pd.DataFrame loaded with all daily rainfall
    
    args:
        lat: integer rounded to 3 decimals
        lon: integer rounded to 3 decimals
        dataset = SUPPORTED_DATASETS[n] where n is the index of the dataset you want to use
    """
    
    rainfall_dict = get_rainfall_dict(lat, lon, dataset)
    rainfall_dict = {"DATE": [k for k in rainfall_dict.keys()], "PRCP": [k for k in rainfall_dict.values()]}
    
    rainfall_df = pd.DataFrame.from_dict(rainfall_dict)
    rainfall_df.DATE = pd.to_datetime(rainfall_df.DATE)
    
    return rainfall_df.set_index(['DATE'])


def get_temperature_df(lat, lon, dataset_revision):
    """
    Get full temperature data from one of the temperature datasets
    Args:
        lat: float to 3 decimals
        lon: float to 3 decimals
        dataset_revision: the name of the dataset as listed on the ipfs gateway
    Returns:
        a pandas DataFrame with cols DATE, HIGH and LOW
    """
    highs, lows = get_temperature_dict(lat, lon, dataset_revision)
    intermediate_dict = {
        "DATE": [date for date in highs],
        "HIGH": [highs[date] for date in highs],
        "LOW": [lows[date] for date in lows]
    }
    temperature_df = pd.DataFrame.from_dict(intermediate_dict)
    temperature_df.DATE = pd.to_datetime(temperature_df.DATE)

    return  temperature_df.set_index(["DATE"])


def get_station_df(station_id):
    """ Get a given station's raw data as a pandas dataframe. """
    df = pd.read_csv(io.StringIO(get_station_csv(station_id)))
    return df.set_index(pd.DatetimeIndex(df['DATE']))


def get_station_rainfall_df(station_id):
    """ Get full daily rainfall time series from GHCN Station Data. """
    return get_station_df(station_id)[['PRCP', 'NAME']]
    
    
def get_station_temperature_df(station_id):
    """ Get full daily min, max temp time series from GHCN Station Data. """
    
    return get_station_df(station_id)[['TMIN', 'TMAX', 'NAME']]


def get_station_snow_df(station_id):
    """ Get full daily snowfall time series from GHCN Station Data in mm. """
    return get_station_df(station_id)[['SNOW', 'NAME']]




"""
Some functions for organizing dWeather query results into dataframe formats relevant to Arbol's use cases.
Mainly related to summing and averaging over time.
This module imports pandas and numpy, the arithmetic.py module should not
"""
import numpy as np
import pandas as pd
from dweather_client.http_client import get_rainfall_dict, get_rev_rainfall_dict, get_heads, get_metadata
from dweather_client.utils import listify_period, get_n_closest_station_ids
from dweather_client.ipfs_client import cat_station_df, cat_station_csv, cat_metadata
from dweather_client.df_utils import get_station_ids_with_icao
import ipfshttpclient
import datetime
import logging

HISTORICAL_START_YEAR = 1981

def cat_station_df_list(station_ids, pin=True, force_hash=None):
    batch_hash = force_hash
    if (force_hash is None):
        batch_hash = get_heads()['ghcnd']
    metadata = cat_metadata(batch_hash, pin=pin)
    station_content = []
    with ipfshttpclient.connect() as client:
        for station_id in station_ids:
            logging.info("(%i of %i): Loading station %s from %s into DataFrame%s" % ( \
                station_ids.index(station_id),
                len(station_ids),
                station_id, 
                "dWeather head" if force_hash is None else "forced hash",
                " and pinning to ipfs datastore" if pin else ""
            ))
            station_content.append(cat_station_df( \
                station_id,
                client=client,
                pin=pin,
                force_hash=batch_hash
            ))
    return station_content 

def cat_icao_stations(pin=True, force_hash=None):
    """
    For every station that has an icao code, load it into a dataframe and
    return them all as a list.
    """
    station_ids = get_station_ids_with_icao()
    return cat_station_df_list(station_ids, pin=pin, force_hash=force_hash)

def cat_n_closest_station_dfs(lat, lon, n, pin=True, force_hash=None):
    """
    Load the closest n stations to a given point into a list of dataframes.
    """
    if (force_hash is None):
        metadata = cat_metadata(get_heads()['ghcnd'])
    else:
        metadata = cat_metadata(force_hash)
    station_ids = get_n_closest_station_ids(lat, lon, metadata, n)
    return cat_station_df_list(station_ids, pin=pin, force_hash=force_hash)

def sum_period_df(df, ps, pe, yrs, peril):
    """ 
    Get the sums of a peril by a defined period.
    
    return:
        pd.DataFrame indexed by year
        
    args:
        df = weather data with a daily index
        ps = contract start period, datetime.date object
        pe = contract end period, datetime.date object
        yrs = years lookback
        peril = the name of column you wish to sum
        
    """
    newdf = df
    years = np.array([])
    sums = np.array([])

    for year in range(ps.year-yrs, ps.year):
        if ps.year < pe.year: # if period crosses jan 1 into a new year
            left = str(year)+'-'+str(ps.month)+'-'+str(ps.day)
            right = str(year+1)+'-'+str(pe.month)+'-'+str(pe.day)
        else:
            left = str(year)+'-'+str(ps.month)+'-'+str(ps.day)
            right = str(year)+'-'+str(pe.month)+'-'+str(pe.day)

        sumChunk = newdf[left:right]
        years = np.append(years, int(year))
        sums = np.append(sums, sumChunk[peril].sum())
    return pd.DataFrame({'year': years, 'value': sums}).set_index("year")


def avg_period_df(df, ps, pe, yrs, peril):
    """ 
    Gets the avg of a peril by a defined period
    
    return:
        pd.DataFrame indexed by year
        
    args:
        df = weather data with a daily index
        ps = contract start period, datetime.date object
        pe = contract end period, datetime.date object
        yrs = years lookback
        peril = the name of column you wish to avg
        
    """

    newdf = df
    years = np.array([])
    avgs = np.array([])

    for year in range(ps.year-yrs, ps.year):
        if ps.year < pe.year: # if period crosses jan 1 into a new year
            left = str(year)+'-'+str(ps.month)+'-'+str(ps.day)
            right = str(year+1)+'-'+str(pe.month)+'-'+str(pe.day)
        else:
            left = str(year)+'-'+str(ps.month)+'-'+str(ps.day)
            right = str(year)+'-'+str(pe.month)+'-'+str(pe.day)

        sumChunk = newdf[left:right]
        years = np.append(years,int(year))
        avgs = np.append(avgs, sumChunk[peril].mean())
    return pd.DataFrame({'year': years, 'value': avgs}).set_index('year')



from dweather_client.http_client import get_rainfall_dict, get_temperature_dict, get_station_csv, parse_station_temps_as_dict
import dweather_client.ipfs_datasets

class GridCellDataLoader:

    _instances = {}

    def __init__(self, dataset, lat, lon, weather_type):
        """
        Args:
            lat (float): the grid cell latitude
            lon (float): the grid cell longitude
            weather_type (str): either rainfall or temperature
        """
        self.revisions = {}
        self.dataset = dataset
        self.lat = lat
        self.lon = lon
        assert weather_type == "rainfall" or weather_type == "temperature", \
            "weather_type must be 'rainfall' or 'temperature'"
        self.weather_type = weather_type

    def __new__(cls, dataset, lat, lon, weather_type):
        """ Create new class only if one with args doesn't exist
        Similar to singleton class, we only want one instance of DataLoader for each grid cell
        """
        if (dataset, lat, lon, weather_type) not in cls._instances:
            cls._instances[(dataset, lat, lon, weather_type)] = super(GridCellDataLoader, cls).__new__(cls)
        return cls._instances[(dataset, lat, lon, weather_type)]

    
    def load_revision(self, revision):
        """ Get the revision from ipfs and save it
        Args:
            revision (str): the named dataset revision on the ipfs gateway
        """
        if self.weather_type == "rainfall":
            metadata, data = get_rainfall_dict(self.lat, self.lon, revision, return_metadata=True)
        elif self.weather_type == "temperature":
            metadata, highs, lows = get_temperature_dict(self.lat, self.lon, revision, return_metadata=True)
            data = {"highs": highs, "lows": lows}
        self.revisions[revision] = {"metadata": metadata, "data": data}

    def populate_revision(self, revision):
        """ Fetch the revision only if we don't already have it
        Args:
            revision (str): the revision name on the ipfs gateway
        """
        if revision not in self.revisions:
            self.load_revision(revision)

    def get_revision(self, revision):
        """ Return the saved data revision if it exists, otherwise load it and return it
        Args:
            revisiont (str): the dataset revision name as on the ipfs gateway
        Returns:
            dictionary (str):
                {
                    "metadata": the revision metadata
                    "data": the revision data
                }
            for rainfall, "data" is a dict of datetime.date: rainfall values
            for temperature, "data" is a dict { "highs": dict, "lows": dict} with highs and lows associating datetime.dates with temps
        """
        self.populate_revision(revision)
        return self.revisions[revision]

    def reload(self):
        """ Get the lastest data from all saved revisions. Writes over saved revisions """
        for revision in self.revisions:
            self.load_revision(revision)

    def load_all_revisions(self):
        """ Get all revisions of the dataset. Writes over saved revisions"""
        for revision in dweather_client.ipfs_datasets.datasets[self.dataset]:
            self.load_revision(revision)

    def build_multi_revision_dict(self):
        """ Build out a dictionary using the most accurate data possible 
        Returns:
            dictionary
        """
        if self.weather_type == "rainfall":
            all_data_dict = {}
        elif self.weather_type == "temperature":
            all_data_dict = {
                "highs": {},
                "lows": {}
            }
        for revision in dweather_client.ipfs_datasets.datasets[self.dataset]:
            self.populate_revision(revision)
            if self.weather_type == "rainfall":
                new_data = {date: (self.revisions[revision]['data'][date], revision) for date in self.revisions[revision]["data"] if date not in all_data_dict}
                all_data_dict.update(new_data)
            elif self.weather_type == "temperature":
                new_highs = {date: (self.revisions[revision]['data']['highs'][date], revision) for date in self.revisions[revision]["data"]["highs"] if date not in all_data_dict["highs"]}
                new_lows = {date: (self.revisions[revision]['data']['lows'][date], revision) for date in self.revisions[revision]["data"]["lows"] if date not in all_data_dict["lows"]}
                all_data_dict["highs"].update(new_highs)
                all_data_dict["lows"].update(new_lows)

        return all_data_dict

class StationDataLoader:
    _instances = {}

    def __init__(self, station_id):
        """
        Args:
            station_id (str): the station id
        """
        self.station_id = station_id
        self.csv_text = ""

    def __new__(cls, station_id):
        """ Create new class only if one with args doesn't exist
        Similar to singleton class, we only want one instance of DataLoader for each station
        """
        if station_id not in cls._instances:
            cls._instances[station_id] = super(StationDataLoader, cls).__new__(cls)
        return cls._instances[station_id]

    def populate_csv(self):
        """ Get the revision from ipfs and save it
        Args:
            revision (str): the named dataset revision on the ipfs gateway
        """
        if self.csv_text == '':
            self.csv_text = get_station_csv(self.station_id)

    def get_temperatures(self, use_fahrenheit=True):
        """ Return the station data Tmins and Tmaxs 
        Args:
            use_fahrenheit: use degrees F if true, degrees C if false
        returns:
            tuple of dicts:
                datetime.date: float the daily highs
                datetime.date: float the daily lows
        """
        self.populate_csv()
        tmaxs, tmins = parse_station_temps_as_dict(self.csv_text, use_fahrenheit)
        return {"highs": tmaxs, "lows": tmins}
        




from dweather_client import http_client, utils
from datetime import date

def test_http_client_rainfall():
    dataset = 'chirps_05'
    dataset_revision = 'chirps_05-daily'
    heads = http_client.get_heads()
    chirps_metadata = http_client.get_metadata(heads[dataset_revision])
    rainfall_dict = http_client.get_rainfall_dict(41.175, -75.125, dataset_revision)
    rainfall_rev_dict = http_client.get_rev_rainfall_dict(41.175, -75.125, dataset, date.today(), dataset_revision)

def test_http_client_temperature():
    dataset = "cpc_temp"
    dataset_revision = 'cpc_temp-daily'
    lat, lon = utils.conventional_lat_lon_to_cpc(41.25, -77.75)
    heads = http_client.get_heads()
    cpc_metadata = http_client.get_metadata(heads[dataset_revision])
    temperature_dict = http_client.get_temperature_dict(lat, lon, dataset_revision)
    temperature_rev_dict = http_client.get_rev_temperature_dict(lat, lon, dataset, date.today(), dataset_revision)
    tagged_temperature_rev_dict = http_client.get_rev_tagged_temperature_dict(lat, lon, dataset)




"""
Helpful auxilliary functions that don't directly interact with IPFS.
This module specifically excludes pandas
"""
import math
import datetime
from heapq import heappush, heappushpop
from geopy.distance import geodesic

def snap_to_grid(lat, lon, metadata):
    """ 
    Find the nearest (lat,lon) on IPFS for a given metadata file.
    return: lat, lon
    args:
        lat = -90 < lat < 90, float
        lon = -180 < lon < 180, float
        metadata: a dWeather metadata file

    """
    resolution = metadata['resolution']
    min_lat = metadata['latitude range'][0] #start [lat, lon]
    min_lon = metadata['longitude range'][0] #end [lat, lon]
    category = metadata['climate category']

    if 'cpc' in metadata['source data url']:
        min_lat, min_lon = conventional_lat_lon_to_cpc(min_lat, min_lon)

    # check that the lat lon is in the bounding box
    snap_lat = round(round((lat - min_lat)/resolution) * resolution + min_lat, 3)
    snap_lon = round(round((lon - min_lon)/resolution) * resolution + min_lon, 3)
    return snap_lat, snap_lon

def get_n_closest_station_ids(lat, lon, metadata, n):
    pq = []
    for feature in metadata["stations"]["features"]:
        s_lat = float(feature["geometry"]["coordinates"][0])
        s_lon = float(feature["geometry"]["coordinates"][1])
        distance = geodesic([lat, lon], [s_lat, s_lon]).miles
        station_id = feature["properties"]["station id"]
        if (len(pq) >= n):
            heappushpop(pq, (1 / distance, station_id))
        else:
            heappush(pq, (1 / distance, station_id))
    return [pair[1] for pair in pq]

def cpc_lat_lon_to_conventional(lat, lon):
    """
    Convert a pair of coordinates from the idiosyncratic CPC lat lon
    format to the conventional lat lon format.
    """
    return lat, lon - 360


def conventional_lat_lon_to_cpc(lat, lon):
    """
    Convert a pair of coordinates from conventional (lat,lon)
    to the idiosyncratic CPC (lat,lon) format.
    """
    return lat, lon + 360


def listify_period(start_date, end_date):
    """
    Make a list of all dates from start_date to end_date, inclusive.
    Args:
        start_date (datetime.date): first date to include
        end_date (datetime.date): last date to include
    Returns:
        list of datetime.date objects
    """
    days_in_range = (end_date - start_date).days
    return [start_date + datetime.timedelta(n) for n in range(days_in_range + 1)]


def is_revision_final(dataset, revision_to_check, last_acceptable_revision):
    """
    See if a given dataset revision should be considered 'final'
    a final dataset is any dataset at or before the last acceptable one
    Args:
        dataset (str): the dataset name (e.g. prism_precip)
        revision_to_check (str): the revision tag under consideration for final status
        last_acceptable_revision (str): the last 'final' revision tag
    returns:
        bool true if the revision is considered final
    """
    # The dataset revision lists are ordered by accuracy so we simply compare the indicies in the list
    dataset_list = IPFSDatasets.datasets[dataset]
    return dataset_list.index(revision_to_check) <= dataset_list.index(last_acceptable_revision)

def celcius_to_fahrenheit(deg_c):
    return round((deg_c * 9/5) + 32, 5)




from dweather_client import ipfs_client, http_client, utils
from datetime import date

#def test_ipfs_client_all_stations():
#    ipfs_client.pin_all_stations()

def test_ipfs_client_rainfall():
    dataset = 'chirps_05'
    dataset_revision = 'chirps_05-daily'
    heads = http_client.get_heads()
    chirps_metadata = ipfs_client.cat_metadata(heads[dataset_revision])
    ipfs_client.cat_rainfall_dict(41.175, -75.125, dataset_revision)
    rainfall_rev_dict = ipfs_client.cat_rev_rainfall_dict(41.175, -75.125, dataset, date.today(), dataset_revision)

def test_http_client_temperature():
    dataset = "cpc_temp"
    dataset_revision = 'cpc_temp-daily'
    lat, lon = utils.conventional_lat_lon_to_cpc(41.25, -77.75)
    heads = http_client.get_heads()
    cpc_metadata = ipfs_client.cat_metadata(heads[dataset_revision])
    temperature_dict = ipfs_client.cat_temperature_dict(lat, lon, dataset_revision)
    temperature_rev_dict = ipfs_client.cat_rev_temperature_dict(lat, lon, dataset, date.today(), dataset_revision)
   



import setuptools

try:
    # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # for pip <= 9.0.3
    from pip.req import parse_requirements

def load_requirements(fname):
    reqs = parse_requirements(fname, session="test")
    try:
        return [str(ir.requirement) for ir in reqs]
    except AttributeError:
        return [str(ir.req) for ir in reqs]

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="dweather_client",
    include_package_data=True,
    install_requires=load_requirements("requirements.txt"),
    version="1.4.4",
    author="Arbol",
    author_email="info@arbolmarket.com",
    description="Python client for interacting with weather data on IPFS.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Arbol-Project/dWeather-Python-Client.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Some functions for organizing dWeather query results into formats relevant to Arbol's use cases.
Mainly related to summing and averaging over time.
This module should not import pandas or numpy
"""
from dweather_client.http_client import get_rainfall_dict, get_rev_rainfall_dict
from dweather_client.utils import listify_period
import datetime

HISTORICAL_START_YEAR = 1981

def sum_period_rainfall(lat, lon, dataset, start_date, end_date, daily_cap=None, use_prelim=False, final_rev=None):
    ''' Download the rainfall data from ipfs and compute the rainfall for the given term
    Args:
        lat (float): latitude of grid cell
        lon (float): longitude of grid cell
        dataset (str): the rainfall dataset name in ipfs
        start_date (datetime.date): first date of term to include
        end_date (datetime.date): last date of term to include
        daily_cap (float): max rainfall in mm to count for a day
        use_prelim (bool): if true, include prelim data and get as much of the period as possible
    returns:
        (float, int, bool)
            float: the total rainfall
            int: the number of days in the period
            bool: true if all rainfall is in the 'Final' dataset
    Raises:
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as rainfall data
    '''
    is_final = True
    if use_prelim:
        rainfall_dict, is_final = get_rev_rainfall_dict(lat, lon, dataset, end_date, final_rev)
        # The rainfall dict is sorted by date, so the last date will be the most recent (prelim) data
        # Need to check if the end date is in the rainfall dict, and if not, truncate the term to the end of the data
        if not is_final and end_date not in rainfall_dict:
            end_date = list(rainfall_dict)[-1]
    else:
        rainfall_dict = get_rainfall_dict(lat, lon, dataset)
    dates = listify_period(start_date, end_date)
    if daily_cap:
        rain = [min(rainfall_dict[date], daily_cap) for date in dates]
    else:
        rain = [rainfall_dict[date] for date in dates]
    return (sum(rain), len(rain), is_final)

def build_historical_rainfall_dict(lat, lon, dataset, start_date, end_date, daily_cap=None, start_year=HISTORICAL_START_YEAR, end_year=None, use_prelim=False, final_rev=None, ignore_missing=False):
    ''' Builds a dict of rainfall values over the term period for each year for a grid cell
    Args:
        lat (float): latitude of grid cell
        lon (float): longitude of grid cell
        dataset (str): the rainfall dataset name in ipfs
        start_date (datetime.date): first date of term to include
        end_date (datetime.date): last date of term to include
        daily_cap (float): max rainfall in mm to count for a day
        start_year (int): the first year of historical data to include, using the start date
        end_year (int): the last year of data to include with the start_date year
        use_prelim (bool): if True, include prelim rainfall if there is not enough final rainfall
                if there is not enough rainfall including prelim then truncate the period to where we have data
        ignore_missing (bool): if true, substitute '0' for the missing rainfall value
                if false, set the whole year as None
    returns
        dict of int: (float, int): keys are the start date year for each term, values are
                a tuple of total rainfall and the number of days in the term for the term period starting that year
        bool is_final: true if all data included is from the final dataset
        If there is missing rainfall data for a period, the value for that year will be None, None
    Raises:
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as rainfall data
    '''
    is_final = True
    if use_prelim:
        rainfall_dict, is_final = get_rev_rainfall_dict(lat, lon, dataset, end_date, final_rev)
        # The rainfall dict is sorted by date, so the last date will be the most recent (prelim) data
        # Need to check if the end date is in the rainfall dict, and if not, truncate the term to the end of the data
        if not is_final and end_date not in rainfall_dict:
            end_date = list(rainfall_dict)[-1]
    else:
        rainfall_dict = get_rainfall_dict(lat, lon, dataset)
    yearly_rainfall = {}
    if end_year is None:
        # If the end year is not provided they probably want all data up to the current year
        end_year = start_date.year
    for year in range(start_year, end_year + 1):
        historical_start = start_date.replace(year=year)
        # end date is tricky because it could be in the next calendar year
        historical_end = end_date.replace(year=year + (end_date.year - start_date.year))
        year_term = listify_period(historical_start, historical_end)
        yearly_rain = [rainfall_dict[date] for date in year_term]
        if None in yearly_rain:
            if ignore_missing:
                # replace None values with 0
                yearly_rain = [x if x is not None else 0 for x in yearly_rain] 
            else:
                # return None for the whole year
                yearly_rainfall[year] = None, None
                continue
        if daily_cap:
            yearly_rain = list(map(lambda x: min(x, daily_cap), yearly_rain))
        yearly_rainfall[year] = (sum(yearly_rain), len(yearly_rain))
            
    return yearly_rainfall, is_final



from dweather_client import arithmetic, df_arithmetic, df_loader, utils, df_utils
import datetime

TODAY = datetime.date.today()
ONE_MONTH = datetime.timedelta(days=30)


def test_build_historical_rainfall_dict():
    historical_raindict = arithmetic.build_historical_rainfall_dict(
        41.175, 
        -75.125, 
        'chirps_05-daily', 
        TODAY - (ONE_MONTH * 8),
        TODAY - (ONE_MONTH * 7),
    )


def test_sum_period_rainfall():
    a_sum = arithmetic.sum_period_rainfall(
         41.175, 
        -75.125, 
        'chirps_05-daily', 
        TODAY - (ONE_MONTH * 8),
        TODAY - (ONE_MONTH * 7),
    )


def test_sum_period_df():
    rainfall_df = df_loader.get_rainfall_df(41.125, -75.125, 'chirps_05-daily')
    sliced_period = df_utils.period_slice_df(
        rainfall_df, 
        TODAY - (ONE_MONTH * 8), 
        TODAY - (ONE_MONTH * 7), 
        30,
    )
    df_sum = df_arithmetic.sum_period_df(
        rainfall_df, 
        TODAY - (ONE_MONTH * 8), 
        TODAY - (ONE_MONTH * 7),
        30, 
        "PRCP")


def test_avg_period_df():
    rainfall_df = df_loader.get_rainfall_df(41.125, -75.125, 'chirps_05-daily')
    sliced_period = df_utils.period_slice_df(
        rainfall_df, 
        TODAY - (ONE_MONTH * 8), 
        TODAY - (ONE_MONTH * 7), 
        30,
    )
    df_avg = df_arithmetic.avg_period_df(
        rainfall_df, 
        TODAY - (ONE_MONTH * 8), 
        TODAY - (ONE_MONTH * 7),
        30, 
        "PRCP")



from dweather_client import df_loader, utils
import datetime

def test_rainfall_df():
    rainfall_df = df_loader.get_rainfall_df(41.125, -75.125, 'chirps_05-daily')

def test_temperature_df():
    temp_df = df_loader.get_temperature_df(43.083, -92.000, 'prism_temp-daily')

def test_station_dfs():
    station_df = df_loader.get_station_rainfall_df('USW00024285')
    station_df2 = df_loader.get_station_temperature_df('USW00024285')
    station_df3 = df_loader.get_station_snow_df('USW00024285')



"""
Helpful auxilliary functions that don't directly interact with IPFS.
Basically, it should go here if it doesn't need to import from df_loader or http_client,
this module uses pandas
"""
import math
import pandas as pd
import datetime
import os

PARENT_DIR = os.path.dirname(os.path.abspath(__file__))
CPC_LOOKUP_PATH = PARENT_DIR + '/etc/cpc-grid-ids.csv'
ICAO_LOOKUP_PATH = PARENT_DIR + '/etc/airport-codes.csv'

def dataframeify(dict_or_df):
    """
    Convert a dict to a dataframe.
    If "dict" is already a dataframe, just return it.
    """
    if isinstance(dict_or_df, pd.DataFrame):
        return dict_or_df
    else:
        return pd.DataFrame.from_dict(dict_or_df, orient="index")


def cpc_grid_to_lat_lon(grid):
    """ 
    Convert a cpc grid id to conventional lat lon via a lookup table.
    return:
        latitude, longitude
        
    args:
        grid = "1100" example
    
    """
    cpc_grids =  pd.read_csv(os.path.join(CPC_LOOKUP_PATH)).set_index('Grid ID') #dataframe of cpc grid to lat/lon lookup table
    myGrid = cpc_grids.iloc[grid]
    coords = [myGrid["Latitude"], myGrid["Longitude"]]
    coords = [coord+360 if coord < 0 else coord for coord in coords] #converts negative coordinates to positive values
    
    return coords[0], coords[1]


def get_station_ids_with_icao():
    """
    Get a list of station id that are associated with stations that have an icao code.
    """
    icao_lookup = pd.read_csv(os.path.join(ICAO_LOOKUP_PATH)).set_index('ICAO')
    ids = []
    for index, row in icao_lookup.iterrows():
        if (row["GHCN"] not in ids):
            ids.append(row["GHCN"])
    return ids

def icao_to_ghcn(icao):
    """ 
    Convert an icao airport code to ghcn.
    return:
        latitude, longitude
    args:
        icao = "xxxx" for example try "KLGA"
    """
    icao_codes = pd.read_csv(os.path.join(ICAO_LOOKUP_PATH)).set_index('ICAO') #get lookup table
    return icao_codes.loc[icao]["GHCN"]


def period_slice_df(df, ps, pe, yrs):
    """ 
    Slice a dataframe by period where n is based on years lookback 
    
    return:
        pd.DataFrame()
    
    args:
        df = any pd.DataFrame() with a datetime.date() index
        ps = datetime.date object (contract period start)
        ps = datetime.date object (contract period end)
        yrs = 30 (years lookback)
        
    limitations:
        max period is 364 days
        
    """

    mydf = pd.DataFrame()
    
    leftDay = ~((df.index.day < ps.day) & (df.index.month == ps.month))
    rightDay = ~((df.index.day > pe.day) & (df.index.month == pe.month))
    #first year just get end
    if ps.year < pe.year:
        firstYear = ps.year - yrs
        yr = df.index.year == firstYear
        mon = df.index.month >= ps.month
        #a boolean mask to cut the left and right half month days if need be
        mydf = mydf.append(df[(yr)&(mon)&(leftDay)])
        #everything in betweeng
        for yrss in range(ps.year - (yrs - 1), pe.year - 1):
            ##boolean mask for now, better method needed##
            yearChunk = (df.index.year == yrss)&((df.index.month<=pe.month)|(df.index.month>=ps.month))
            yearChunk = (yearChunk)&leftDay&rightDay
            mydf = mydf.append(df[yearChunk])
        #build last year
        mydf = mydf.append(df[rightDay&(df.index.month<=pe.month)&(df.index.year==pe.year-1)])
    else:
        mydf = df[leftDay&rightDay&(df.index.month>=ps.month)&(df.index.month<=pe.month)&(df.index.year>=ps.year-yrs)]
            
    return mydf




import ipfshttpclient, json, requests, datetime, io, gzip
from dweather_client.ipfs_errors import *
from dweather_client.utils import listify_period
from dweather_client.df_utils import get_station_ids_with_icao
import dweather_client.ipfs_datasets
import pandas as pd

MM_TO_INCHES = 0.0393701
RAINFALL_PRECISION = 5
GATEWAY_URL = 'https://gateway.arbolmarket.com'


def get_heads(url=GATEWAY_URL):
    """
    Get heads.json for a given IPFS gateway.
    Args:
        url (str): base url of the IPFS gateway url
    Returns (example heads.json):
        {
            'chirps_05-daily': 'Qm...',
            'chirps_05-monthly': 'Qm...',
            'chirps_25-daily': 'Qm...',
            'chirps_25-monthly': 'Qm...',
            'cpc_us-daily': 'Qm...',
            'cpc_us-monthly': 'Qm...'
        }
    """
    hashes_url = url + "/climate/hashes/heads.json"
    r = requests.get(hashes_url)
    r.raise_for_status()
    return r.json()


def cat_metadata(hash_str, client=None, pin=True):
    """
    Get the metadata file for a given hash.
    Args:
        url (str): the url of the IPFS server
        hash_str (str): the hash of the ipfs dataset
    Returns (example metadata.json):
    
        {
            'date range': [
                '1981/01/01',
                '2019/07/31'
            ],
            'entry delimiter': ',',
            'latitude range': [
                -49.975, 49.975
            ],
            'longitude range': [
                -179.975, 179.975]
            ,
            'name': 'CHIRPS .05 Daily Full Set Uncompressed',
            'period': 'daily',
            'precision': 0.01,
            'resolution': 0.05,
            'unit of measurement': 'mm',
            'year delimiter': '\n'
        }
    """
    session_client = ipfshttpclient.connect() if client is None else client
    try:
        if pin:
            session_client.pin.add(hash_str + "/metadata.json")
        metadata = session_client.cat(hash_str + "/metadata.json")
    finally: 
        if (client is None):
            session_client.close()
    return json.loads(metadata)


def cat_hash_cell(hash_str, coord_str, client=None):
    if (client is None):
        with ipfshttpclient.connect() as client:
            return client.cat(hash_str + '/' + coord_str)
    else:
        return client.cat(hash_str + '/' + coord_str)

def cat_zipped_hash_cell(url, hash_str, coord_str, client=None):
    """
    Read a text file on the ipfs server compressed with gzip.
    Args:
        url (str): the url of the ipfs server
        hash_str (str): the hash of the dataset
        coord_str (str): the text file coordinate name e.g. 45.000_-96.000
    Returns:
        the contents of the file as a string
    """
    if (client is None):
        with ipfshttpclient.connect() as client:
            cell = client.cat(hash_str + '/' + coord_str + ".gz")
            with gzip.GzipFile(fileobj=io.BytesIO(cell)) as zip_data:
                return zip_data.read().decode("utf-8")
    else:
        cell = client.cat(hash_str + '/' + coord_str + ".gz")
        with gzip.GzipFile(fileobj=io.BytesIO(cell)) as zip_data:
            return zip_data.read().decode("utf-8")


def cat_dataset_cell(lat, lon, dataset_revision, client=None):
    """ 
    Retrieve the text of a grid cell data file for a given lat lon and dataset.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        A tuple (json, str) of the dataset metadata file and the grid cell data text
    Raises: 
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
    """
    all_hashes = get_heads()
    if dataset_revision in all_hashes:
        dataset_hash = all_hashes[dataset_revision]
    else:
        raise DatasetError('{} not found on server'.format(dataset_revision))

    metadata = cat_metadata(dataset_hash, client)
    min_lat, max_lat = sorted(metadata["latitude range"])
    min_lon, max_lon = sorted(metadata["longitude range"])
    if lat < min_lat or lat > max_lat:
        raise InputOutOfRangeError("Latitude {} out of dataset revision range [{:.3f}, {:.3f}] for {}".format(lat, min_lat, max_lat, dataset_revision))
    if  lon < min_lon or lon > max_lon:
        raise InputOutOfRangeError("Longitude {} out of dataset revision range [{:.3f}, {:.3f}] for {}".format(lon, min_lon, max_lon, dataset_revision))
    coord_str = "{:.3f}_{:.3f}".format(lat,lon)
    try:
        if "compression" in metadata and metadata["compression"] == "gzip":
            text_data = cat_zipped_hash_cell(GATEWAY_URL, dataset_hash, coord_str, client=client)
        else:
            text_data = cat_hash_cell(dataset_hash, coord_str, client=client)
        return metadata, text_data
    except requests.exceptions.HTTPError as e:
        raise CoordinateNotFoundError('Coordinate ({}, {}) not found  on ipfs in dataset revision {}'.format(lat, lon, dataset_revision))



def cat_rainfall_dict(lat, lon, dataset_revision, return_metadata=False, client=None):
    """ 
    Build a dict of rainfall data for a given grid cell.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        a dict ({datetime.date: float}) of datetime dates and the corresponding rainfall in mm for that date
    Raises:
        DatasetError: If no matching dataset found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as rainfall data
    """
    metadata, rainfall_text = cat_dataset_cell(lat, lon, dataset_revision, client=client)
    dataset_start_date = datetime.datetime.strptime(metadata['date range'][0], "%Y/%m/%d").date()
    dataset_end_date = datetime.datetime.strptime(metadata['date range'][1], "%Y/%m/%d").date()
    timedelta = dataset_end_date - dataset_start_date
    days_in_record = timedelta.days + 1 # we have both the start and end date in the dataset so its the difference + 1
    try:
        rainfall_text = rainfall_text.decode()
    except:
        pass
    day_strs = rainfall_text.replace(',', ' ').split()
    if (len(day_strs) != days_in_record):
        raise DataMalformedError ("Number of days in data file does not match the provided metadata")
    rainfall_dict = {}
    for i in range(days_in_record):
        if day_strs[i] == metadata["missing value"]:
            rainfall_dict[dataset_start_date + datetime.timedelta(days=i)] = None
        else:
            rainfall_dict[dataset_start_date + datetime.timedelta(days=i)] = float(day_strs[i])
    if return_metadata:
        return metadata, rainfall_dict
    else:
        return rainfall_dict
  

def cat_rev_rainfall_dict(lat, lon, dataset, desired_end_date, latest_rev):
    """
    Build a dictionary of rainfall data. Include as much of the most accurate, final data as possible. Start by buidling from the most accurate data,
    then keep appending data from more recent/less accurate versions of the dataset until we run out or reach the end date.

    This will not throw an error if there are no revisions with data available, it will simply return what is available.
    Args:
        lat (float): the grid cell latitude
        lon (float): the grid cell longitude
        dataset (str): the name of the dataset, e.g., "chirps_05-daily" on hashes.json
        desired_end_date (datetime.date): the last day of data needed.
        latest_rev (str): the least accurate revision of the dataset that is considered final
    Returns:
        tuple:
            a dict ({datetime.date: float}) of datetime dates and the corresponding rainfall in mm for that date
            bool is_final: if all data up to desired end date is final, this will be true
    """
    all_rainfall = {}
    is_final = True
    with ipfshttpclient.connect() as client:
        # Build the rainfall from the most accurate revision of the dataset to the least
        for dataset_revision in dweather_client.ipfs_datasets.datasets[dataset]:
            additional_rainfall = cat_rainfall_dict(lat, lon, dataset_revision, client=client)
            all_dates = list(all_rainfall) + list(additional_rainfall)
            all_rainfall = {date: all_rainfall[date] if date in all_rainfall else additional_rainfall[date] for date in all_dates}
            # stop when we have the desired end date in the dataset
            if desired_end_date in all_rainfall:
                return all_rainfall, is_final
            # data is no longer final after we pass the specified version
            if dataset_revision == latest_rev:
                is_final = False

    # If we don't reach the desired dataset, return all data.
    return all_rainfall, is_final


def cat_temperature_dict(lat, lon, dataset_revision, return_metadata=False, client=None):
    """
    Build a dict of temperature data for a given grid cell.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
    Returns:
        tuple (highs, lows) of dicts
        highs: dict ({datetime.date: float}) of datetime dates and the corresponding high temperature in degress F
        lows: dict ({datetime.date: float}) of datetime dates and the corresponding low temperature in degress F
    Raises:
        DatasetError: If no matching dataset_revision found on server
        InputOutOfRangeError: If the lat/lon is outside the dataset_revision range in metadata
        CoordinateNotFoundError: If the lat/lon coordinate is not found on server
        DataMalformedError: If the grid cell file can't be parsed as temperature data
    """
    metadata, temp_text = cat_dataset_cell(lat, lon, dataset_revision, client=client)
    dataset_start_date = datetime.datetime.strptime(metadata['date range'][0], "%Y/%m/%d").date()
    dataset_end_date = datetime.datetime.strptime(metadata['date range'][1], "%Y/%m/%d").date()
    timedelta = dataset_end_date - dataset_start_date
    days_in_record = timedelta.days + 1 # we have both the start and end date in the dataset_revision so its the difference + 1
    try:
        temp_text = temp_text.decode()
    except:
        pass
    day_strs = temp_text.replace(',', ' ').split()
    if (len(day_strs) != days_in_record):
        raise DataMalformedError ("Number of days in data file does not match the provided metadata")
    highs = {}
    lows = {}
    for i in range(days_in_record):
        low, high = map(float, day_strs[i].split('/'))
        date_iter = dataset_start_date + datetime.timedelta(days=i)
        highs[date_iter] = high
        lows[date_iter] = low
    if return_metadata:
        return metadata, highs, lows
    else:
        return highs, lows


def cat_rev_temperature_dict(lat, lon, dataset, desired_end_date, latest_rev):
    """
    Build a dictionary of rainfall data. Include as much final data as possible. If the desired end date
    is not in the final dataset, append as much prelim as possible.
    Args:
        lat (float): the latitude of the grid cell, to 3 decimals
        lon (float): the longitude of the grid cell, to 3 decimals
        dataset (str): the dataset name as on hashes.json
        desired_end_date (datetime.date): don't include prelim data after this point if not needed
        latest_rev (str): The least accurate revision that is still considered 'final'
    returns:
        tuple (highs, lows) of dicts and a bool
        highs: dict ({datetime.date: float}) of datetime dates and the corresponding high temperature in degress F
        lows: dict ({datetime.date: float}) of datetime dates and the corresponding low temperature in degress F
        is_final: True if all data is from final dataset, false if prelim included
    """
    highs = {}
    lows = {}
    is_final = True

    with ipfshttpclient.connect() as client:
        # Build the data from the most accurate version of the dataset to the least
        for dataset_revision in dweather_client.ipfs_datasets.datasets[dataset]:
            additional_highs, additional_lows = cat_temperature_dict(lat, lon, dataset_revision, client=client)
            all_dates = list(highs) + list(additional_highs)    
            highs = {date: highs[date] if date in highs else additional_highs[date] for date in all_dates}
            lows = {date: lows[date] if date in lows else additional_lows[date] for date in all_dates}
            # Stop early if we have the end date
            if desired_end_date in highs:
                return highs, lows, is_final

            # data is no longer final after we pass the specified version
            if dataset_revision == latest_rev:
                is_final = False

    # If we don't reach the desired dataset, return all data.
    return highs, lows, is_final


def pin_all_stations(client=None):
    """ Sync all stations locally."""
    heads = get_heads()
    dataset_hash = heads["ghcnd"]
    session_client = ipfshttpclient.connect() if client is None else client
    try:
        session_client.pin.add(dataset_hash)
    finally:
        if (client is None):
            session_client.close()

def cat_station_df(station_id, client=None, pin=True, force_hash=None):
    """ Cat a given station's raw data as a pandas dataframe. """
    df = pd.read_csv(io.StringIO(\
        cat_station_csv(
            station_id, 
            client=client, 
            pin=pin,
            force_hash=force_hash
        )
    ))
    return df.set_index(pd.DatetimeIndex(df['DATE']))

def cat_station_csv(station_id, client=None, pin=True, force_hash=None):
    """
    Retrieve the contents of a station data csv file.
    Args:
        station_id (str): the id of the weather station
    returns:
        the contents of the station csv file as a string
    """
    if (force_hash is None):
        all_hashes = get_heads()
        dataset_hash = all_hashes["ghcnd"]
    else:
        dataset_hash = force_hash
    csv_hash = dataset_hash + '/' + station_id + ".csv.gz"
    session_client = ipfshttpclient.connect() if client is None else client
    try:
        if pin:
            session_client.pin.add(csv_hash)
        csv = session_client.cat(csv_hash)
        with gzip.GzipFile(fileobj=io.BytesIO(csv)) as zip_data:
            return zip_data.read().decode("utf-8")
    finally:
        if (client is None):
            session_client.close()


def cat_icao_stations(client=None, pin=True):
    """ Get a list of station dataframes for all stations that have an icao"""
    dfs = []
    session_client = ipfshttpclient.connect() if client is None else client
    try:
        for station_id in get_station_ids_with_icao():
            try:
                print(station_id)
                dfs.append(cat_station_csv(station_id, client=client, pin=pin))
            except Exception as e:
                print(e)
                continue
    finally:
        if (client is None):
            session_client.close()
    return dfs
